class PromptBuilder {
  String clientId;
  final Map<String, dynamic> _nodes = {};
  final Map<String, Map<String, dynamic>> _outputToNode = {};
  int _nodeIdCounter = 1;

  PromptBuilder({required this.clientId});
  Map<String, dynamic> toJson() {
    return {
      "nodes": _nodes,
      "outputToNode": _outputToNode,
      "nodeIdCounter": _nodeIdCounter
    };
  }

  void fromJson(Map<String, dynamic> json) {
    _nodes.clear();
    _outputToNode.clear();
    _nodeIdCounter = json["nodeIdCounter"];

    for (var entry in json["nodes"].entries) {
      _nodes[entry.key] = entry.value;
    }

    for (var entry in json["outputToNode"].entries) {
      _outputToNode[entry.key] = entry.value;
    }
  }

  /// Adds a node to the workflow and returns its nodeId
  String addNode(String classType, Map<String, dynamic> inputs,
      {String? title, Map<String, String>? outputTags}) {
    final nodeId = _nodeIdCounter.toString();

    _nodes[nodeId] = {
      "inputs": inputs,
      "class_type": classType,
      "_meta": {"title": title ?? classType}
    };

    // Register outputs of this node with optional tags
    final defaultOutputs = _getDefaultOutputs(classType);
    for (var i = 0; i < defaultOutputs.length; i++) {
      final outputTag = outputTags?[defaultOutputs[i]] ?? defaultOutputs[i];
      _outputToNode[outputTag] = {
        "nodeId": nodeId,
        "outputIndex": i.toString()
      };
    }

    _nodeIdCounter++;
    return nodeId;
  }

  /// Provides access to the current nodes
  Map<String, dynamic> get nodes => Map.unmodifiable(_nodes);

  /// Edits an existing node in the workflow
  void editNode(String nodeId,
      {Map<String, dynamic>? newInputs, String? newTitle}) {
    if (!_nodes.containsKey(nodeId)) {
      throw Exception("Node with ID $nodeId does not exist.");
    }

    final node = _nodes[nodeId];
    if (newInputs != null) {
      node["inputs"] = newInputs; // Inputs will be resolved in build()
    }

    if (newTitle != null) {
      node["_meta"]["title"] = newTitle;
    }
  }

  /// Removes a node from the workflow
  void removeNode(String nodeId) {
    if (!_nodes.containsKey(nodeId)) {
      throw Exception("Node with ID $nodeId does not exist.");
    }

    // Remove the node
    _nodes.remove(nodeId);

    // Remove associated outputs
    _outputToNode.removeWhere((_, value) => value["nodeId"] == nodeId);
  }

  /// Reorders nodes in the workflow
  void reorderNodes(List<String> newOrder) {
    if (newOrder.length != _nodes.keys.length ||
        !newOrder.every((id) => _nodes.containsKey(id))) {
      throw Exception("Invalid new order: All node IDs must be included.");
    }

    final reorderedNodes = <String, dynamic>{};
    for (var nodeId in newOrder) {
      reorderedNodes[nodeId] = _nodes[nodeId];
    }
    _nodes
      ..clear()
      ..addAll(reorderedNodes);
  }

  /// Inserts a node between an existing node's output and its consumers.
  ///
  /// - [classType]: The class type of the node to insert.
  /// - [inputs]: The specific inputs for the new node (excluding connections being intercepted).
  /// - [targetNodeId]: The ID of the node *before* the insertion point.
  /// - [targetOutputIndices]: The output indices of the [targetNodeId] to intercept (e.g., [0, 1] for MODEL/CLIP).
  /// - [newNodeOutputTags]: A map defining the symbolic tags for the *new* node's outputs, keyed by the output name (e.g., {"MODEL": "new_model_tag", "CLIP": "new_clip_tag"}).
  /// - [title]: Optional title for the new node.
  ///
  /// Returns the ID of the newly inserted node.
  String insertNode({
    required String classType,
    required Map<String, dynamic> inputs,
    required String targetNodeId,
    required List<int> targetOutputIndices,
    required Map<String, String> newNodeOutputTags,
    String? title,
  }) {
    if (!_nodes.containsKey(targetNodeId)) {
      throw Exception("Target node with ID $targetNodeId does not exist.");
    }
    print("[insertNode] Inserting $classType after target $targetNodeId"); // Log 1

    // 1. Find original symbolic tags for the target node's outputs being intercepted
    final Map<int, String> originalTargetTags = {}; // { outputIndex: tag }
    _outputToNode.forEach((tag, info) {
      if (info['nodeId'] == targetNodeId &&
          targetOutputIndices.contains(int.parse(info['outputIndex']))) {
         originalTargetTags[int.parse(info['outputIndex'])] = tag;
       }
     });
    print("[insertNode] Found original target tags: $originalTargetTags"); // Log 2

    if (originalTargetTags.length != targetOutputIndices.length) {
      throw Exception(
          "Could not find all specified original output tags for target node $targetNodeId.");
    }

    // 2. Find consumers connected to these original tags
    final Map<String, Map<String, int>> consumers = {}; // { consumerNodeId: { inputKey: originalTargetOutputIndex } }
    print("[insertNode] Searching for consumers connected to tags: ${originalTargetTags.values}"); // Log 3
    _nodes.forEach((consumerNodeId, consumerNodeData) {
      final consumerInputs = consumerNodeData['inputs'] as Map<String, dynamic>? ?? {};
      consumerInputs.forEach((inputKey, inputValue) {
        if (inputValue is List && inputValue.length == 2 && inputValue[0] is String) {
          final sourceTag = inputValue[0] as String;
          final sourceIndex = inputValue[1] as int;
          // Check if this input is connected to one of the tags we are intercepting
          // Added check for null safety on originalTargetTags[sourceIndex]
          if (originalTargetTags.containsKey(sourceIndex) && originalTargetTags[sourceIndex] == sourceTag) {
             consumers.putIfAbsent(consumerNodeId, () => {})[inputKey] = sourceIndex;
          }
         }
       });
     });
    print("[insertNode] Found consumers: $consumers"); // Log 4

    // 3. Prepare inputs for the new node (combining provided inputs and intercepted connections)
    final Map<String, dynamic> newNodeInputs = Map.from(inputs);
    originalTargetTags.forEach((outputIndex, tag) {
      // Determine the input key for the new node based on the output type (assuming MODEL/CLIP convention)
      // This might need refinement if inserting other node types
      final inputKey = (outputIndex == 0) ? 'model' : (outputIndex == 1) ? 'clip' : null;
      if (inputKey != null) {
        newNodeInputs[inputKey] = [tag, outputIndex];
      } else {
         print("Warning: Could not determine input key for intercepted output index $outputIndex");
      }
    });
    print("[insertNode] Prepared new node inputs: $newNodeInputs"); // Log 5


    // 4. Add the new node using the existing addNode logic
    // Note: We need to ensure the outputTags passed to addNode match the structure it expects (keyed by default output name)
    final defaultOutputs = _getDefaultOutputs(classType); // e.g., ["MODEL", "CLIP"]
    final Map<String, String> internalOutputTags = {};
    newNodeOutputTags.forEach((key, value) {
       if (defaultOutputs.contains(key)) {
         internalOutputTags[key] = value;
       }
    });
    print("[insertNode] Adding node with internalOutputTags: $internalOutputTags"); // Log 6

    final newNodeId = addNode(
      classType,
      newNodeInputs,
      title: title,
      outputTags: internalOutputTags, // Pass the structured tags
    );
    print("[insertNode] Added node with ID: $newNodeId"); // Log 7

    // 5. Rewire consumers to point to the new node's output tags
    print("[insertNode] Starting consumer rewiring..."); // Log 8
    consumers.forEach((consumerNodeId, inputMap) {
      print("[insertNode] Rewiring consumer: $consumerNodeId"); // Log 9
      inputMap.forEach((inputKey, originalTargetOutputIndex) {
        print("[insertNode]   Input: $inputKey (original index: $originalTargetOutputIndex)"); // Log 10
        // Find the corresponding *new* output tag based on the original index
        final outputName = (originalTargetOutputIndex == 0) ? "MODEL" : (originalTargetOutputIndex == 1) ? "CLIP" : null;
        if (outputName != null && newNodeOutputTags.containsKey(outputName)) {
            final newTag = newNodeOutputTags[outputName]!;
            final newConnection = [newTag, originalTargetOutputIndex];
            print("[insertNode]     New connection: $newConnection (using tag '$newTag')"); // Log 11

            // Get the consumer's current inputs
            final currentConsumerInputs = Map<String, dynamic>.from(
                _nodes[consumerNodeId]?['inputs'] ?? {});
            print("[insertNode]     Current inputs before edit: $currentConsumerInputs"); // Log 12 (Changed name slightly)

            // Update only the specific input being rewired
            currentConsumerInputs[inputKey] = newConnection;
            print("[insertNode]     Updated inputs map: $currentConsumerInputs"); // Log 13 (Changed name slightly)

            // Use editNode with the full, updated inputs map
            editNode(consumerNodeId, newInputs: currentConsumerInputs);
            print("[insertNode]     Called editNode for $consumerNodeId"); // Log 14
         } else {
            print("Warning: Could not find new output tag or determine output name for consumer $consumerNodeId.$inputKey (original index: $originalTargetOutputIndex)");
        }
      });
    });
    print("[insertNode] Finished consumer rewiring."); // Log 15

    return newNodeId;
  }


  /// Generates the final workflow map
  Map<String, dynamic> build() {
    final resolvedNodes = <String, dynamic>{};

    _nodes.forEach((nodeId, node) {
      final resolvedInputs = <String, dynamic>{};

      // Resolve dependencies for inputs
      node["inputs"].forEach((key, value) {
        if (value is List && value.length == 2 && value[1] is int) {
          final outputTag = value[0];
          if (_outputToNode.containsKey(outputTag)) {
            resolvedInputs[key] = [
              _outputToNode[outputTag]!["nodeId"],
              value[1]
            ];
          } else {
            throw Exception(
                "Unresolved dependency: No node provides output tag '$outputTag'");
          }
        } else {
          resolvedInputs[key] = value;
        }
      });

      resolvedNodes[nodeId] = {
        ...node,
        "inputs": resolvedInputs,
      };
    });

    return {
      "client_id": clientId,
      "prompt": resolvedNodes,
    };
  }

  /// Validates the workflow against object info (optional)
  void validate(Map<String, dynamic> objectInfo) {
    for (var nodeId in _nodes.keys) {
      final node = _nodes[nodeId];
      final classType = node["class_type"];
      if (!objectInfo.containsKey(classType)) {
        throw Exception("Invalid node class type: $classType");
      }
      final requiredInputs = objectInfo[classType]["input"]["required"];
      for (var inputKey in requiredInputs.keys) {
        if (!node["inputs"].containsKey(inputKey)) {
          throw Exception(
              "Missing required input '$inputKey' for node $nodeId");
        }
      }
    }
  }

  /// Helper method to get default outputs for a class type
  List<String> _getDefaultOutputs(String classType) {
    switch (classType) {
      case "KSampler":
        return ["LATENT"];
      case "CheckpointLoaderSimple":
        return ["MODEL", "CLIP", "VAE"];
      case "CLIPTextEncode":
        return ["CONDITIONING"];
      case "CLIPSetLastLayer":
        return ["CLIP"];
      case "VAEDecode":
        return ["IMAGE"];
      case "VAEEncode":
        return ["LATENT"];
      case "VAEEncodeForInpaint":
        return ["LATENT"];
      case "VAELoader":
        return ["VAE"];
      case "EmptyLatentImage":
        return ["LATENT"];
      case "EmptySD3LatentImage":
        return ["LATENT"];
      case "LatentUpscale":
        return ["LATENT"];
      case "LatentUpscaleBy":
        return ["LATENT"];
      case "LatentFromBatch":
        return ["LATENT"];
      case "RepeatLatentBatch":
        return ["LATENT"];
      case "SaveImage":
        return [];
      case "PreviewImage":
        return [];
      case "LoadImage":
        return ["IMAGE", "MASK"];
      case "LoadImageMask":
        return ["MASK"];
      case "LoadImageOutput":
        return ["IMAGE", "MASK"];
      case "ImageScale":
        return ["IMAGE"];
      case "ImageScaleBy":
        return ["IMAGE"];
      case "ImageInvert":
        return ["IMAGE"];
      case "ImageBatch":
        return ["IMAGE"];
      case "ImagePadForOutpaint":
        return ["IMAGE", "MASK"];
      case "EmptyImage":
        return ["IMAGE"];
      case "ConditioningAverage":
        return ["CONDITIONING"];
      case "ConditioningCombine":
        return ["CONDITIONING"];
      case "ConditioningConcat":
        return ["CONDITIONING"];
      case "ConditioningSetArea":
        return ["CONDITIONING"];
      case "ConditioningSetAreaPercentage":
        return ["CONDITIONING"];
      case "ConditioningSetAreaStrength":
        return ["CONDITIONING"];
      case "ConditioningSetMask":
        return ["CONDITIONING"];
      case "KSamplerAdvanced":
        return ["LATENT"];
      case "SetLatentNoiseMask":
        return ["LATENT"];
      case "LatentComposite":
        return ["LATENT"];
      case "LatentBlend":
        return ["LATENT"];
      case "LatentRotate":
        return ["LATENT"];
      case "LatentFlip":
        return ["LATENT"];
      case "LatentCrop":
        return ["LATENT"];
      case "LoraLoader":
        return ["MODEL", "CLIP"];
      case "CLIPLoader":
        return ["CLIP"];
      case "UNETLoader":
        return ["MODEL"];
      case "DualCLIPLoader":
        return ["CLIP"];
      case "CLIPVisionEncode":
        return ["CLIP_VISION_OUTPUT"];
      case "StyleModelApply":
        return ["CONDITIONING"];
      case "StyleModelLoader":
        return ["STYLE_MODEL"];
      case "CLIPVisionLoader":
        return ["CLIP_VISION"];
      case "VAEDecodeTiled":
        return ["IMAGE"];
      case "VAEEncodeTiled":
        return ["LATENT"];
      case "unCLIPCheckpointLoader":
        return ["MODEL", "CLIP", "VAE", "CLIP_VISION"];
      case "GLIGENLoader":
        return ["GLIGEN"];
      case "GLIGENTextBoxApply":
        return ["CONDITIONING"];
      case "InpaintModelConditioning":
        return ["positive", "negative", "latent"];
      case "LatentAdd":
        return ["LATENT"];
      case "LatentSubtract":
        return ["LATENT"];
      case "LatentMultiply":
        return ["LATENT"];
      case "LatentInterpolate":
        return ["LATENT"];
      case "LatentBatch":
        return ["LATENT"];
      case "LatentBatchSeedBehavior":
        return ["LATENT"];
      case "LatentApplyOperation":
        return ["LATENT"];
      case "LatentApplyOperationCFG":
        return ["MODEL"];
      case "LatentOperationTonemapReinhard":
        return ["LATENT_OPERATION"];
      case "LatentOperationSharpen":
        return ["LATENT_OPERATION"];
      case "HypernetworkLoader":
        return ["MODEL"];
      case "UpscaleModelLoader":
        return ["UPSCALE_MODEL"];
      case "ImageUpscaleWithModel":
        return ["IMAGE"];
      case "ImageBlend":
        return ["IMAGE"];
      case "ImageBlur":
        return ["IMAGE"];
      case "ImageQuantize":
        return ["IMAGE"];
      case "ImageSharpen":
        return ["IMAGE"];
      case "ImageScaleToTotalPixels":
        return ["IMAGE"];
      case "LatentCompositeMasked":
        return ["LATENT"];
      case "ImageCompositeMasked":
        return ["IMAGE"];
      case "MaskToImage":
        return ["IMAGE"];
      case "ImageToMask":
        return ["MASK"];
      case "ImageColorToMask":
        return ["MASK"];
      case "SolidMask":
        return ["MASK"];
      case "InvertMask":
        return ["MASK"];
      case "CropMask":
        return ["MASK"];
      case "MaskComposite":
        return ["MASK"];
      case "FeatherMask":
        return ["MASK"];
      case "GrowMask":
        return ["MASK"];
      case "ThresholdMask":
        return ["MASK"];
      case "PorterDuffImageComposite":
        return ["IMAGE", "MASK"];
      case "SplitImageWithAlpha":
        return ["IMAGE", "MASK"];
      case "JoinImageWithAlpha":
        return ["IMAGE"];
      case "RebatchLatents":
        return ["LATENT"];
      case "RebatchImages":
        return ["IMAGE"];
      case "ModelMergeSimple":
        return ["MODEL"];
      case "ModelMergeBlocks":
        return ["MODEL"];
      case "ModelMergeSubtract":
        return ["MODEL"];
      case "ModelMergeAdd":
        return ["MODEL"];
      case "CheckpointSave":
        return [];
      case "CLIPSave":
        return [];
      case "VAESave":
        return [];
      case "ModelSave":
        return [];
      case "TomePatchModel":
        return ["MODEL"];
      case "CLIPTextEncodeSDXLRefiner":
        return ["CONDITIONING"];
      case "CLIPTextEncodeSDXL":
        return ["CONDITIONING"];
      case "Canny":
        return ["IMAGE"];
      case "FreeU":
        return ["MODEL"];
      case "FreeU_V2":
        return ["MODEL"];
      case "SamplerCustom":
        return ["LATENT", "LATENT"];
      case "BasicScheduler":
        return ["SIGMAS"];
      case "KarrasScheduler":
        return ["SIGMAS"];
      case "ExponentialScheduler":
        return ["SIGMAS"];
      case "PolyexponentialScheduler":
        return ["SIGMAS"];
      case "LaplaceScheduler":
        return ["SIGMAS"];
      case "VPScheduler":
        return ["SIGMAS"];
      case "BetaSamplingScheduler":
        return ["SIGMAS"];
      case "SDTurboScheduler":
        return ["SIGMAS"];
      case "KSamplerSelect":
        return ["SAMPLER"];
      case "SamplerEulerAncestral":
        return ["SAMPLER"];
      case "SamplerEulerAncestralCFGPP":
        return ["SAMPLER"];
      case "SamplerLMS":
        return ["SAMPLER"];
      case "SamplerDPMPP_3M_SDE":
        return ["SAMPLER"];
      case "SamplerDPMPP_2M_SDE":
        return ["SAMPLER"];
      case "SamplerDPMPP_SDE":
        return ["SAMPLER"];
      case "SamplerDPMPP_2S_Ancestral":
        return ["SAMPLER"];
      case "SamplerDPMAdaptative":
        return ["SAMPLER"];
      case "SplitSigmas":
        return ["SIGMAS", "SIGMAS"];
      case "SplitSigmasDenoise":
        return ["SIGMAS", "SIGMAS"];
      case "FlipSigmas":
        return ["SIGMAS"];
      case "SetFirstSigma":
        return ["SIGMAS"];
      case "CFGGuider":
        return ["GUIDER"];
      default:
        return [];
    }
  }
}
